"""unique email contraint

Revision ID: b2e694dfde35
Revises: 89913992dd5a
Create Date: 2018-05-17 09:14:01.087221

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = 'b2e694dfde35'
down_revision = '89913992dd5a'


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('ix_address_email', table_name='address')
    op.create_index(
        op.f('ix_address_email'), 'address', ['email'], unique=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_address_email'), table_name='address')
    op.create_index('ix_address_email', 'address', ['email'], unique=False)
    # ### end Alembic commands ###
